###################################################################################################################
#### Simulation Program, Repeated Measures Design 
#### Part 1: Some Auxiliary Functions
#### Part 2: Simulation - Input
#### Part 3: Simulation 
#### Part 4: Output of Results
###################################################################################################################

rm(list=ls())

library(car) #for traditional within ANOVA
library(statmod) # to create values from inverse Gaussian distribution (invgauss)
library(moments) # for skewness and kurtosis
library(grDevices) # pdf-file: est density of errors

####################### Part 1: Start Auxiliary Functions ########################################################

### Creates matrix D_i (see text)
### Input: x.1 - matrix whose rows are to be replicated, entries in vector x.2 give the no. of replications  
rip = function(x.1,x.2){
  x.1 = as.matrix(x.1)
  x.2 = as.matrix(x.2)
  stopifnot(ncol(x.2)==1)
  out = apply(x.1,2,rep,x.2)
  return(out)
}

### Horicontal direct product of two matrices x.1 and x.2: direct product on all rows of x.1 and x.2, 
### Result is a new matrix by stacking these row vectors into a matrix.
hdir = function(x.1,x.2){ 
  x.1 = as.matrix(x.1)
  x.2 = as.matrix(x.2)
  vers = 0
  if(nrow(x.1)==nrow(x.2)){vers = 1}
  if(nrow(x.2)==1){vers = 2}  
  if(vers==1){
    er = as.matrix(x.1[,rep(seq_len(ncol(x.1)),each=ncol(x.2))])*as.matrix(x.2[,rep.int(seq_len(ncol(x.2)),ncol(x.1))])
  }
  if(vers==2){
    er = x.1%x%matrix(x.2,nrow=1) 
  }
  if(vers==0){
    er = NA
    print("ERROR in hdir: Matrices have not same number of rows")
  }else{
    colnames(er) = paste(rep(colnames(x.1),each=ncol(x.2)),rep.int(colnames(x.2),ncol(x.1)),sep="*")}
  return(er)
}

### Generates design matrix for a saturated design, asymmetric or symmetric restrictions  
### Input: K - vector, legth(K)= no. of factors, whose elements give the no. of factor levels
###        res: 0 - asymmetric restrictions, 
###             1 - symmetric restrictions
###             otherwise identity matrix
###        Fames: names of elements in K 
m.eff = function(K,ia=0,res=1,Fnames=NA){
  x = c()
  cond = c()
  if(res==0){
    if(length(Fnames)==1){if(is.na(Fnames)){Fnames = LETTERS[1:length(K)]}}
    ### asymmetric restrictions
    for(k in 1:length(K)){
      if(k==1 & length(K)==1){x = cbind(x,diag(K)[,-1]); cond = matrix(seq_len(K),ncol=1); 
      colnames(x) = paste(Fnames[1],1:(K[1]-1),sep=".")}    
      if(k==1 & length(K)>1){
        x = kronecker(diag(K[k])[,-1],matrix(1,prod(K[-k]),1))
        colnames(x) = paste(Fnames[1],1:(K[1]-1),sep=".")
        cond = kronecker(seq_len(K[1]),matrix(1,prod(K[-1]),1))} 
      if(k>1  & k<length(K)){
        temp = kronecker(matrix(1,prod(K[1:(k-1)]),1),diag(K[k])[,-1])
        temp = kronecker(temp,matrix(1,prod(K[(k+1):length(K)]),1))
        colnames(temp) = paste(Fnames[k],1:(K[k]-1),sep=".")
        if(ia==0){x = cbind(x,temp)}else{x = cbind(x,temp,hdir(x,temp))}
        cond = cbind(cond,kronecker(
          kronecker(matrix(1,prod(K[1:(k-1)]),1),seq_len(K[k])),
          matrix(1,prod(K[(k+1):length(K)]),1)))}
      if(k>1  & k==length(K)){
        temp = kronecker(matrix(1,prod(K[1:(length(K)-1)]),1),diag(K[k])[,-1])
        colnames(temp) = paste(Fnames[k],1:(K[k]-1),sep=".")        
        if(ia==0){x = cbind(x,temp)}else{x = cbind(x,temp,hdir(x,temp))}
        cond = cbind(cond,kronecker(matrix(1,prod(K[1:(length(K)-1)]),1),seq_len(K[k])))          
      }
    }
  }
  if(res==1){
    if(length(Fnames)==1){if(is.na(Fnames)){Fnames = LETTERS[1:length(K)]}}    
    ### symmetric restrictions
    for(k in 1:length(K)){
      if(k==1 & length(K)==1){
        temp = as.matrix(diag(K[1])[,-K[1]])
        temp[nrow(temp),] = -1
        x = cbind(x,temp) 
        cond = matrix(seq_len(K[1]),ncol=1)
        colnames(x) = paste(Fnames[1],1:(K[1]-1),sep=".")}    
      if(k==1 & length(K)>1){
        x = as.matrix(diag(K[k])[,-K[k]])
        x[nrow(x),] = -1
        x = kronecker(x,matrix(1,prod(K[-k]),1))
        colnames(x) = paste(Fnames[1],1:(K[1]-1),sep=".")
        cond = kronecker(seq_len(K[1]),matrix(1,prod(K[-1]),1))} 
      if(k>1  & k<length(K)){
        temp = as.matrix(diag(K[k])[,-K[k]])
        temp[nrow(temp),] = -1
        temp = kronecker(matrix(1,prod(K[1:(k-1)]),1),temp)
        temp = kronecker(temp,matrix(1,prod(K[(k+1):length(K)]),1))
        colnames(temp) = paste(Fnames[k],1:(K[k]-1),sep=".")
        if(ia==0){x = cbind(x,temp)}else{x = cbind(x,temp,hdir(x,temp))}
        cond = cbind(cond,kronecker(
          kronecker(matrix(1,prod(K[1:(k-1)]),1),seq_len(K[k])),
          matrix(1,prod(K[(k+1):length(K)]),1)))}
      if(k>1 & k==length(K)){
        temp = as.matrix(diag(K[k])[,-K[k]])
        temp[nrow(temp),] = -1
        temp = kronecker(matrix(1,prod(K[1:(length(K)-1)]),1),temp)
        colnames(temp) = paste(Fnames[k],1:(K[k]-1),sep=".")
        if(ia==0){x = cbind(x,temp)}else{x = cbind(x,temp,hdir(x,temp))}
        cond = cbind(cond,kronecker(matrix(1,prod(K[1:(length(K)-1)]),1),seq_len(K[k])))
      }
    }
  }
  if(res!=0 & res!=1){x = diag(prod(K))}else{x = cbind(mu=1,x)}
  return(list(x=x,cond=cond))
} 

### create matrices C for main and interacton effects tests
### Input: K - vector, legth(K)= no. of factors, whose elements give the no. of factor levels
###        Fames: names of elements in K 
gen.C = function(K,ia=1,Fnames=NA){
  if(length(Fnames)==1){if(is.na(Fnames)){Fnames = LETTERS[1:length(K)]}}
  if(ia==1){
    pos.xx = matrix(1,2,1)
    nms = c("mu")    
    for(k in seq_len(length(K))){
      nms = c(nms,Fnames[k])
      tmp = matrix(c(pos.xx[2,ncol(pos.xx)]+1,pos.xx[2,ncol(pos.xx)]+K[k]-1),2,1)
      pos.xx = cbind(pos.xx,tmp)
      if(k > 1){
        for(l in 2:(ncol(pos.xx)-1)){
          tmp.0 = (pos.xx[2,l] - pos.xx[1,l] + 1) * (tmp[2] - tmp[1] + 1)
          pos.xx = cbind(pos.xx,matrix(c(pos.xx[2,ncol(pos.xx)]+1,pos.xx[2,ncol(pos.xx)]+tmp.0),2,1))
          nms = c(nms,paste(nms[l],":",Fnames[k],sep=""))
          }
      }
    }
    tmp = NULL
    tmp[paste("C.",1:ncol(pos.xx),sep="")] = list(NULL)
    tmp[[1]] = matrix(c(1,rep(0,prod(K)-1)),1,prod(K))
    for(l in 2:length(tmp)){
      nr = pos.xx[2,l] - pos.xx[1,l] + 1
      nc = pos.xx[1,l] -1
      if(l < length(tmp)){tmp[[l]] = cbind(matrix(0,nr,nc),diag(nr),matrix(0,nr,prod(K)-nc-nr))}else{
        tmp[[l]] = cbind(matrix(0,nr,nc),diag(nr))}
    names(tmp) = nms
    }
  }else{tmp = NULL}
  return(tmp)
}

### Calculate acceleration parameter for BCa confidence interval 
accel = function(dat){apply(as.matrix(dat),2,function(x){sum((mean(x)-x)**3)/(6*sum((mean(x)-x)**2)**(3/2))})}

############################### End Auxiliary Functions ########################################################


####################### Part 2: Simulation - Input #############################################################

set.seed(5575553) # set starting value for (pseudo-) random number generators

alpha = 0.05  # confidence interval: nominal coverage (1 - alpha)
nsim  = 500   # no. of simulations
n = 33        # no. of units
#n = 200 

# stopping criteria for iterative GEE estimation
max.it = 40     # maximum no. of iterations
eps = 1e-6      # maximum absolute difference between estimates 

## factors, factor levels and true theta as adopted in the main text
K = c(4,2,2,2)  # vector of factor levels, no. of elements = no. of factors
theta = round(rnorm(prod(K)),2) # true effects (theta); generated randomly from N(0,1) distribution

## Factors, factor levels and true theta for comparison with Anova in supplement - START
#K = c(3,5)
#theta = rep(0,prod(K)) # true effects (theta)

GEE.iter = 1 # calculate simplified iterative GEE only if GEE.iter==1
rep.anov = 0 # repeated measures ANOVA only if rep.anov==1

dist.re = c(0,0,0)
dist.re = c(4,3,4) # distribution of random effects an: 0 = normal    
#dist.re = c(3,3,3)#                                    1 = t-dist., df=fg 
                   #                                    2 = chi^2-dist, df=fg
                   #                                    3 = inverse Gauss
                   #                                    4 = "negative" inverse Gauss

# main text: dist.re = c(0,0,0) and dist.re = c(4,3,4)
# supplementary text: dist.re = c(0,0,0) and dist.re = c(3,3,3)

dist.p = c(fg=3,mu=1,lambda=.2)  # fg = df for t- and chi^2-distribution, mu and lambda for 
                                 # inverse Gause distribution

s2.v = round(runif(prod(K),1,4),2) # error variances varying over conditions; for main text
#s2.v = round(runif(prod(K),16,127),2) # error variances varying over conditions; for supplementary material

# high correlation condition, rho.1 and rho.2 as described in the paper
rho.1 = .5 
rho.2 = .8 

# low correlation condition
#rho.1 = .25 
#rho.2 = .3

# times: no. of observations wihin cells
# p.T.ik: probability/probabilities 
#   if times is a scalar (>0) and p.T.ik=1, then this is the number of observations in all cells 
#      for all units
#   if times is a scalar (>0) and p.T.ik is a vector with numbers larger than 0 but smaller than 1, 
#      then the no. of observations within cells is randomly selected with probabilities given in p.T.ik 
#      from 1 to the value in times for each unit
#   if the no. of elemens in times is larger than one, then the no. of observations for each unit in each cell is
#      randomly selected from times ignoring p.T.ik
# Note: The no. of elements in p.T.ik should either be one or equal to the no. of cells 
#      (otherwise the entries are recycled if applicable)

times = 16

p.T.ik = runif(prod(K),min=.5,max=1)
#summary(p.T.ik)

n.boot1 = 999 # see MacKinnon, 2002
n.boot2 = 999

## create factor levels (= cond), 
## regular design matrix, asymmetric restrictions (= Z); used for the main paper and the real data example
Z = m.eff(K,ia=1,res=0)
## regular design matrix, symmetric restrictions (= Z); used for the comparison with Anova
#Z = m.eff(K,ia=1,res=1)

cond = Z$cond
Z   = Z$x
invZ = solve(Z) # inverse of Z

#### various matrices C for linear combinations of theta, see paper Table 3 and 4
#### create a list, C.mat, consisting of all C-matrices of interest

# if design is like for real data set
if(prod(K)==32){
C.mat = list(
C_1 = matrix(c(0, 1, 0, 0,0,0,0,0,0,0,0,0,0,0,0,0,
               0, 0, 1, 0,0,0,0,0,0,0,0,0,0,0,0,0,
               0, 0, 0, 1,0,0,0,0,0,0,0,0,0,0,0,0),3,16,byrow=TRUE)
,
C_2 = matrix(c(0, 1, 0, 0,0,0,0,0,0,0,0,0,0,0,0,0,
               0,-1, 1, 0,0,0,0,0,0,0,0,0,0,0,0,0,
               0, 0,-1, 1,0,0,0,0,0,0,0,0,0,0,0,0),3,16,byrow=TRUE)
,
C_3 = matrix(c(0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
               0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,
               0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,
               0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,
               0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,
               0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,
               0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,
               0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,
               0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,
               0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0),10,16,byrow=TRUE)
,
C_4 = diag(32)[2:32,]
)

C.mat = lapply(C.mat,function(x){cbind(x,matrix(0,nrow(x),prod(K)-ncol(x)))}) # add a matrix of zeros to each entry in C.mat such that
                                                                              # the corresponding C-matrix has prod(K) columns
}

# if design is NOT like for real data set A - see below for B; NOT USED
if(prod(K)!=32 & prod(K)>2 & length(K)>2){C.mat = list(
  C_1 = cbind(matrix(0,K[1]-1,1), diag(K[1]-1), matrix(0,K[1]-1,prod(K)-K[1])),
  C_2 = cbind(matrix(0,K[2]-1,K[1]), diag(K[2]-1), matrix(0,K[2]-1,prod(K[1:2]-1))),
  C_3 = cbind(matrix(0,prod(K[1:2]-1),1+sum(K[1:2]-1)), diag(prod(K[1:2]-1))),
  C_4 = diag(prod(K))[2:prod(K),])}

if(prod(K)==2){C.mat = list(C_1 = matrix(c(0,1),1,2))} # NOT USED

# if design is NOT like for real data set B; supplementary material
if(rep.anov == 1 & prod(K)>2 & length(K)==2){
  tmp = m.eff(K,ia=1,res=1)
  tmp = solve(tmp$x)
  C.mat = list(
    C_A = tmp[2:K[1],]%*%Z,
    C_B = tmp[(K[1]+1):(sum(K-1)+1),]%*%Z,
    C_AB = tmp[(sum(K-1)+2):ncol(tmp),]%*%Z,
    C_4 = diag(prod(K)))
  }
  
##### value to be tested for (null hypothesis and confidence intervals)
### 'true' value
zeta = lapply(C.mat,function(x)x%*%theta) 
### can be zero 
if(rep.anov == 1){test.zeta = lapply(C.mat,function(x) matrix(0,nrow(x),1)) }
### or
if(rep.anov == 0){test.zeta = zeta}

### variables to collect simulation results
mean.time = 0 # output during simulations: mean time needed for one simulation

est.zeta = as.list(setNames(rep(0,length(C.mat)),paste("zeta",names(C.mat))))

######## Output variables
skew = 0 # output: mean skewness of error distribution
kurt = 0 # output: mean excess kurtosis of error distribution

### standard estimation: Repeated Measures ANOVA on cell menas, Sphericity test and Greenhouse-Geisser correction
res.stand = 0 # F-statistic, sphericity test statistic, corrections 
overall.F = 0 # Overall F-test

### Simple Estimation
out.b = c(); out.v = 0; 
out.ci = 0; out.ci.l = 0; out.ci.u = 0  # confidence intervals under normality assumption
Wald = as.list(setNames(rep(0,length(C.mat)),paste("Wald",names(C.mat))))

### Iterative GEE estimation
out.iter.b = c(); out.iter.v = 0
est.corr = 0 # mean estimated correlations 
est.var = 0  # mean estimated variances
cover.iter.b = 0 # coverage, iterative GEE estimator

### parameters for BC_a bootstrap - only for simple GEE
### coverage
ci.LoUp.a = as.list(setNames(rep(0,length(C.mat)),paste("ci.LoUp.a",names(C.mat))))
LoUp.a = as.list(setNames(rep(0,length(C.mat)),paste("LoUp.a",names(C.mat))))  
out.ci.a = as.list(setNames(rep(0,length(C.mat)),paste("out.ci.a",names(C.mat))))

Wald.b = as.list(setNames(rep(0,length(C.mat)),paste("Wald.b",names(C.mat))))
bca.lin.n.b = as.list(setNames(rep(0,length(C.mat)),paste("bca.lin.n.b",names(C.mat))))
bca.std.n.b = 0 
bca.tmp.LoUp = 0 
LoUp.b = as.list(setNames(rep(0,length(C.mat)),paste("LoUp.b",names(C.mat))))  
bca.lin.ones.n.b = as.list(setNames(rep(0,length(C.mat)),paste("bca.lin.ones.n.b",names(C.mat))))
bca.std.n.ones.b = 0

####################### End Simulation - Input #################################################################

####################### Part 3: Simulation #####################################################################

for(s in seq_len(nsim)){
  time.old = Sys.time()

  if(length(times)==1){
    if(all(p.T.ik==1)){
      T.ik = rep(times,n*nrow(Z))
      if(s==nsim){print("---> No. of observations for each unit and cell identical (balanced design)!")}}
    if(all(0 < p.T.ik) & all(p.T.ik < 1)){
      T.ik = rbinom(n*nrow(Z),times-1,prob=p.T.ik); T.ik = T.ik +1 # n*length(p.T.ik) = nrow(Z) !! 
      if(s==nsim){print("---> If no. of entries in p.T.ik>0: (Possibly) varying selection probabilities in p.T.ik realized!")}}
  }else{
    T.ik = rep(times,times=n)    
    if(s==nsim){print("---> Possible varying selection probabilities in p.T.ik ignored!")}
  }    
  
  T.ik = matrix(T.ik,n,nrow(Z),byrow=TRUE)  # (n x K) 
  T.ik.v = matrix(t(T.ik),n*nrow(Z),1) # (nK x 1)
  
  no.p.n = rowSums(T.ik)  # no. of observations for each subject

  ### distribution of random error components, see Simulation - Input  
  if(dist.re[1]==0){y = rep(rnorm(n,0,sqrt(rho.1)),no.p.n)}
  if(dist.re[1]==1){y = rep(rt(n,df=dist.p["fg"])/sqrt(dist.p["fg"]/(dist.p["fg"]-2))*sqrt(rho.1),no.p.n)} 
  if(dist.re[1]==2){y = rep((rchisq(n,df=dist.p["fg"])-dist.p["fg"])/sqrt(2*dist.p["fg"])*sqrt(rho.1),no.p.n)}
  if(dist.re[1]==3){y = rep((rinvgauss(n,dist.p["mu"],dist.p["lambda"])-dist.p["mu"])/
                             sqrt(dist.p["mu"]**3/dist.p["lambda"])*sqrt(rho.1),no.p.n)}  
  if(dist.re[1]==4){y = rep(-(rinvgauss(n,dist.p["mu"],dist.p["lambda"])-dist.p["mu"])/
                              sqrt(dist.p["mu"]**3/dist.p["lambda"])*sqrt(rho.1),no.p.n)}  
  
  if(dist.re[2]==0){y = y + rep(rnorm(n*prod(K),0,sqrt(rho.2-rho.1)),T.ik.v)}
  if(dist.re[2]==1){y = y + rep(rt(n*prod(K),df=dist.p["fg"])/sqrt(dist.p["fg"]/(dist.p["fg"]-2))*sqrt(rho.2-rho.1),T.ik.v)} 
  if(dist.re[2]==2){y = y + rep((rchisq(n*prod(K),df=dist.p["fg"])-dist.p["fg"])/sqrt(2*dist.p["fg"]) * sqrt(rho.2-rho.1),T.ik.v)}
  if(dist.re[2]==3){y = y + rep((rinvgauss(n*prod(K),dist.p["mu"],dist.p["lambda"])-dist.p["mu"])/
                                sqrt(dist.p["mu"]**3/dist.p["lambda"]) * sqrt(rho.2-rho.1),T.ik.v)}
  if(dist.re[2]==4){y = y + rep(-(rinvgauss(n*prod(K),dist.p["mu"],dist.p["lambda"])-dist.p["mu"])/
                                  sqrt(dist.p["mu"]**3/dist.p["lambda"]) * sqrt(rho.2-rho.1),T.ik.v)}
  
  if(dist.re[3]==0){y = y + rnorm(sum(T.ik),0,sqrt(1-rho.2))}
  if(dist.re[3]==1){y = y + rt(sum(T.ik),df=dist.p["fg"])/sqrt(dist.p["fg"]/(dist.p["fg"]-2))*sqrt(1-rho.2)} 
  if(dist.re[3]==2){y = y + (rchisq(sum(T.ik),df=dist.p["fg"])-dist.p["fg"])/sqrt(2*dist.p["fg"]) * sqrt(1-rho.2)}
  if(dist.re[3]==3){y = y + (rinvgauss(sum(T.ik),dist.p["mu"],dist.p["lambda"])-dist.p["mu"])/
                             sqrt(dist.p["mu"]**3/dist.p["lambda"]) * sqrt(1-rho.2)}  
  if(dist.re[3]==4){y = y + -(rinvgauss(sum(T.ik),dist.p["mu"],dist.p["lambda"])-dist.p["mu"])/
                             sqrt(dist.p["mu"]**3/dist.p["lambda"]) * sqrt(1-rho.2)}  

  if(length(s2.v)==prod(K)){y = rep(rep(sqrt(s2.v),n),T.ik.v)*y} # if variances are varying only over K
  if(length(s2.v)==n*prod(K)){y = rep(sqrt(s2.v),T.ik.v)*y} # if variances vary over n and K
  
  ### calculate skewness and curtosis of errors ignoring dependencies
  skew = (skew*(s-1) + skewness(y))/s
  kurt = (kurt*(s-1) + kurtosis(y))/s

  # the following lines of code print out a graphic of the smoothed density of errors in simulation s (here: s = nsim)
  if(s==nsim){
     # set the path in pdf(...), the name given to the file is 
     # "res-<no. of units>-<no. of outer bootstrap samples>-<rho_2>dist<vector of distributions of random effects>.pdf" 
#    pdf(file=paste("/some directory/out-",n,"-",n.boot1,"-",rho.2,"dist",dist.re,".pdf",sep=""),
#        family="Times",width=6,height=6) #,paper="a4r"
    plot(density(y),main=paste("Estimated density of errors"),
         xlab=expression(paste(sigma[k],v[paste(i,",",k,",",t)])),ylab="",col="black",lwd=1.3)
#    dev.off()     
    }

  y = rep(rep(Z%*%theta,n),T.ik.v)+y
  cond.n = rip(rip(cond,n),T.ik.v)
  indx1 = do.call(`paste0`,as.data.frame(cond.n))
  indx2 = do.call(`paste0`,as.data.frame(cond))
  indx2 = unname(sapply(indx1,function(x) (1:nrow(cond))[indx2==x])) # factor levels for all observations
  rm(indx1)
  id = rep(1:n,no.p.n) # unit identification no. 

  ###################################################################################
  ##### START: standard repeated measures ANOVA
  ##### cell means and balanced design

  if(rep.anov == 1){

    # generate data set tmp.dat to be used in Anova 
    tmp.dat.y = matrix(tapply(y,list(indx2,id),mean),n*prod(K))  # calculate cell means for each i
    cat.pred =  apply(matrix(1,n,1)%x%cond,2,factor) # declare factor-variables

    tmp.dat = as.data.frame(cbind(rep(1:n,each=prod(K)),cat.pred)) # create data frame with id and factor variables; dim: n*prod(K) x no. of factors
    nms = LETTERS[1:ncol(cond.n)]
    names(tmp.dat) = c("id",nms)
    tmp.dat$y = tmp.dat.y 
    rm(tmp.dat.y,cat.pred)

    ## Univariate main and interaction effects, sphericity tests, GG and HF corrections
    t.dat.new = t(matrix(tmp.dat$y,prod(K),n)) # dim: n x K
    nms = sapply(1:ncol(cond),function(x)paste(LETTERS[x],cond[,x],sep=""))
    nms = sapply(1:nrow(nms),function(x) paste(nms[x,],collapse="."))
    colnames(t.dat.new) = nms
    m.1 = lm(t.dat.new ~ 1)

    nms.idata = LETTERS[1:length(K)]
    idata = vector("list", length(nms.idata))   
    names(idata) = nms.idata
    rm(nms.idata)
    for(j in 1:length(K)){
      if(j==1){idata[[1]] = factor(rep(paste(LETTERS[1],"",1:K[1],sep=""),each=prod(K[-1])))}
      if(1<j & j<length(K)){tmp = rep(paste(LETTERS[j],"",1:K[j],sep=""),each=prod(K[-(1:j)]))
                            idata[[j]] = factor(rep(tmp,times=prod(K[1:(j-1)])))}
      if(j==length(K)){idata[[j]] = factor(rep(paste(LETTERS[j],"",1:K[j],sep=""),times=prod(K[-j])))}
      }
    idata = data.frame(idata)
    fmla = as.formula(paste("~",paste(colnames(idata),collapse="*")))
    m.2 = Anova(m.1,idata=idata,idesign= fmla, type=3)
    oldw <- getOption("warn") # temporarily turn off warnings 
                              # (in the simulations these are: HF-epsilon larger 1)
    options(warn = -1)
    m.2 = summary(m.2, multivariate=FALSE)   # attributes(m.2); str(m.2)
    options(warn = oldw)
    # note that multivariate=FALSE does not mean that tests are based on a standard analysis of variance model with independent errors
    # instead it is assumed that sphericity holds and then F-tests from repeated measures analysis of variance are conducted
    # df of within factor with I levels: (I-1); residual sum of squares: interaction of random effect and within factor, df = (I-1)*(n-1)

    tmp.sel = m.2$sphericity.tests[,"p-value"]
    # note: sphericity test and corrections only if no. of factor levels > 2
    tmp = NULL
    tmp$p = m.2$univariate.tests[-1,"Pr(>F)"]
    tmp$"p[GG]" = tmp$p
    tmp$"p[HF]" = tmp$p
    tmp$"Sphericity" = rep(NA,length(tmp$p))
    names(tmp$"Sphericity") = names(tmp$p)
    tmp$"Sphericity"[names(tmp.sel)] = 1
    
    correct = (tmp.sel < alpha) & (tmp$p < alpha) # F test and sphericity test significant?
    if(any(correct)){
      correct = names(correct)[correct]
      tmp$"p[GG]"[correct] = m.2$pval.adjustments[correct,"Pr(>F[GG])"]
      tmp$"p[HF]"[correct] = m.2$pval.adjustments[correct,"Pr(>F[HF])"]
    }
    tmp$"Sphericity" = tmp.sel
    tmp = data.frame("no correction"=m.2$univariate.tests[-1,"Pr(>F)"], # ignore intercept
                     "Sphericity test"=tmp$"Sphericity",
                     "GG correction"=tmp$"p[GG]",
                     "HF correction"=tmp$"p[HF]")
    tmp = apply(tmp>alpha,2,function(x) ifelse(x==TRUE,1,0))
    
    ## collect results on main and interaction effects 
    res.stand = (res.stand*(s-1) + tmp)/s

    ############### F-test for model fit
    #### treat all factor level combinations as levels of one factor
    idata = data.frame(A=factor(1:prod(K)))
    #idata = data.frame(A=factor(cond[,1]))
    m.2 = Anova(m.1,idata=idata,idesign= ~ A, type=3, contrasts="contr.treatment")
    m.2 = summary(m.2, multivariate=FALSE)   # attributes(m.2); str(m.2)
    tmp = c("no correction"=m.2$univariate.tests[-1,"Pr(>F)"],
            "Sphericity test"=m.2$sphericity.tests[,"p-value"],
            "GG correction"=m.2$pval.adjustments[,"Pr(>F[GG])"],
            "HF correction"=m.2$pval.adjustments[,"Pr(>F[HF])"])
    tmp = ifelse((tmp>alpha) ==TRUE,1,0)    
    overall.F = (overall.F *(s-1) + tmp)/s
    rm(tmp,nms,tmp.dat,t.dat.new,idata,m.1,m.2,correct)    
  }
  ##### END: standard repeated measures ANOVA
  ###################################################################################
  
  ###################################################################################
  # START Estimation of simple GEE estimator, see Section 3.2 of the paper 

  mean.y = tapply(y,list(indx2,id),mean) # K x n
  est.theta = invZ%*%rowMeans(mean.y) # calculate estimate for theta, sf. m.1  
  out.b = rbind(out.b,t(est.theta))  # collect theta estimates over simulations
  tmp.v = mean.y - matrix(Z%*%est.theta,ncol(Z),n)  # calculate variance estimate
  rm(mean.y)
  tmp.v = tmp.v%*%t(tmp.v)/(n-1)  # calculate variance estimate (cont.)

  tmp.v = invZ%*%tmp.v%*%solve(t(Z))/n # calculate variance estimate (cont.)
  out.v  = (out.v*(s-1) + diag(tmp.v))/s # calculate variance of theta estimates over simulations
  
  # calculate two-sided symmetric confidence intervals for single parameters assuming normality 
  temp = ((est.theta-sqrt(diag(tmp.v))*qnorm(1-alpha/2)) < theta) & (theta < (est.theta-sqrt(diag(tmp.v))*qnorm(alpha/2)))
  out.ci = (out.ci * (s-1) + temp)/s # calculate actual coverage over simulations

  # calculate one-sided confidence intervals for single parameters assuming normality - not part of the paper
  temp = (theta < (est.theta+sqrt(diag(tmp.v))*qnorm(1-alpha)))
  out.ci.l = (out.ci.l * (s-1) + temp)/s
  temp = ((est.theta-sqrt(diag(tmp.v))*qnorm(1-alpha)) < theta)
  out.ci.u = (out.ci.u * (s-1) + temp)/s
  rm(temp)

  # Standard Wald test: linear hypotheses
  stand.W = lapply(seq_len(length(C.mat)),function(x){
    tmp = solve(C.mat[[x]]%*%tmp.v%*%t(C.mat[[x]]))
    tmp = t(C.mat[[x]]%*%est.theta - test.zeta[[x]])%*%tmp%*%(C.mat[[x]]%*%est.theta - test.zeta[[x]])})
  
  Wald = lapply(seq_len(length(C.mat)),function(x){
                (Wald[[x]]*(s-1) + as.numeric(stand.W[[x]] < qchisq(1-alpha,df=nrow(C.mat[[x]]))))/s})

  # linear combinations: calculate one- and two-sided symmetric confidence intervals assuming normality 
  est.zeta = lapply(seq_len(length(C.mat)),function(x){(est.zeta[[x]]*(s-1)+C.mat[[x]]%*%est.theta)/s})

  ## collect confidence interval results under normality assumption
  # collect coverage rate over simulations for ci of linear combinations of theta, one-sided
  ci.LoUp.a = lapply(seq_len(length(C.mat)),function(x){
    CI.a = sqrt(diag(C.mat[[x]]%*%tmp.v%*%t(C.mat[[x]])))
    tmp.Lo = C.mat[[x]]%*%est.theta-CI.a*qnorm(alpha)
    tmp.Up = C.mat[[x]]%*%est.theta-CI.a*qnorm(1-alpha)
    (ci.LoUp.a[[x]]*(s-1)+cbind(test.zeta[[x]] < tmp.Lo, tmp.Up < test.zeta[[x]]))/s 
  })

  # collect coverage rate over simulations for ci of linear combinations of theta, two-sided, not part of the paper
  # LoUp.a: borders of two-sided confidence intervals, collect over simulations
  CI.a = lapply(seq_len(length(C.mat)),function(x){
                tmp = sqrt(diag(C.mat[[x]]%*%tmp.v%*%t(C.mat[[x]])))
                tmp = cbind(C.mat[[x]]%*%est.theta - tmp * qnorm(1-alpha/2),C.mat[[x]]%*%est.theta - tmp * qnorm(alpha/2))})
  LoUp.a = lapply(seq_len(length(C.mat)),function(x){(LoUp.a[[x]]*(s-1)+CI.a[[x]])/s})
  # collect coverage rate over simulations for two-sided ci of linear combinations of theta
  out.ci.a = lapply(seq_len(length(C.mat)),function(x)
                    (out.ci.a[[x]]*(s-1)+ as.numeric(CI.a[[x]][,1] < test.zeta[[x]] & test.zeta[[x]] < CI.a[[x]][,2]))/s)
  rm(CI.a)
  
  # END Estimation of simple GEE estimator, see Section 3.2 of the paper 
  ###################################################################################

  ###################################################################################
  # START Estimation of standard GEE estimator with simplifications, see Section 3.1 
  if(GEE.iter==1){
    #### Start ## Iterations #########################################
    iter = 0; krit = 0
    est.theta.1 = est.theta # use estimated theta from simple GEE as starting value
    while(iter < max.it & krit==0){
      iter = iter+1
      est.theta.old = est.theta.1
      temp.y = Z%*%est.theta.1 # estimate cell means
      temp.y = y - as.vector(temp.y[indx2]) # calculate residuals for variance and correlation estimation
  
      variance = tapply(temp.y,indx2,function(x) sum(x**2)) # K variances
      variance = variance/colSums(T.ik) # variance estimates

      # calculate correlation estimates, see paper for details
      temp.y = temp.y*as.vector(sqrt(variance[indx2]))
      sr.1 = 0; n.1 = 0
      sr.2 = 0; n.2 = 0
      for(k in seq_len(prod(K))){
        temp.nr = id[indx2==k]
        sr.2 = sr.2 + sum(tapply(temp.y[indx2==k],temp.nr,
                             function(x) sum((x%o%x)*(matrix(1,length(x),length(x))-diag(length(x))))/2))
        n.2 = n.2 + variance[k]**2 * sum(T.ik[,k]*(T.ik[,k]-1)/2) 
        if(k < prod(K)){
          temp.0 = temp.y[indx2==k]
          temp.1 = temp.y[indx2>k]
          ttmp.nr = id[indx2>k]  
          tempr.1 = 0
          tempn.1 = 0
          for(i in seq_len(n)){
            tempr.1 = tempr.1 + sum(temp.0[temp.nr==i]%o%temp.1[ttmp.nr==i])
            temp = indx2[id==i & indx2>k]
            temp = tapply(temp,temp,function(x) length(x))
            tempn.1 = tempn.1 + sum((variance[k]*variance[k<(seq_len(length(variance)))])*
                                    (length(temp.0[temp.nr==i])*temp))
          }
          rm(temp,temp.0,temp.1,ttmp.nr)
        }      
        sr.1 = sr.1 + tempr.1
        n.1  = n.1 + tempn.1
        rm(temp.nr)
      }
      sr.2 = sr.2[[1]]/n.2 
      sr.1 = sr.1/n.1
      rm(n.1,n.2,temp.y)
    
      # calculate inverse of working correlation matrix 
      d.ik = T.ik/(1 - sr.2 + (sr.2 - sr.1)*T.ik) # d_{i,k} as defined in section 3.1, dimension (n x K) 
      m.diag = t(d.ik)  # = matrix(t(d.ik),max(indx2),n), dim: (K x n)
      part.1 = rowSums(m.diag) # = sum_i vec(diag(d_{i,1}, ... , d_{i,K})))
      part.2 = m.diag%*%(t(m.diag)*sr.1/(1+sr.1* colSums(m.diag))) # 
      tD.invR.D = diag(part.1)-part.2 #  inverse of correlation matrix
  
      V = 1/sqrt(variance)
      mean.y = tapply(y,list(indx2,id),mean) # K x n
      mean.y = as.vector(V)*mean.y # (K x n)
      part.1 = rowSums(m.diag*mean.y) # = sum_i vec(diag(d_{i,1}, ... , d_{i,K})))
      part.2 = colSums((t(m.diag)*sr.1/(1+sr.1*colSums(m.diag)))*colSums(m.diag*mean.y))
      tD.invR.y = part.1-part.2
  
      est.theta.1 = solve(t(Z)%*%diag(V)%*%tD.invR.D%*%diag(V)%*%Z)%*%t(Z)%*%diag(V)%*%tD.invR.y
  
      rm(tD.invR.y,part.1,part.2,mean.y)
   
      if(max(abs(est.theta.1-est.theta.old)) < eps){krit = 1}
  
      if(krit == 1){# variance estimation, see estimator (11) in section 3.1
        tmp.v.1 = tapply(y,list(id,indx2),mean) # (n x K)
        tmp.v.1 = tmp.v.1 - matrix(Z%*%est.theta.1,n,ncol(tmp.v.1),byrow=TRUE)
        tmp.v.1 = t(tmp.v.1%*%diag(V)) # (K x n); entspricht V^{-1/2} s^*_i
      
        #m.diag = d.ik  # (n X K)
        m.diag = t(d.ik)*tmp.v.1 # (K x n) 
        part.2 = colSums(m.diag)*sr.1/(1+sr.1*rowSums(d.ik)) # (1 x n)
        part.2 = t(d.ik)%*%diag(part.2) # (K x n)
        part.2 = m.diag - part.2 # (K x n)
        part.2 = part.2%*%t(part.2)
        part.2 = t(Z)%*%diag(V)%*%part.2%*%diag(V)%*%Z    
    
        part.1 = solve(t(Z)%*%diag(V)%*%tD.invR.D%*%diag(V)%*%Z)
        tmp.v.1 = part.1%*%part.2%*%part.1
    
        rm(m.diag,part.1,part.2)
      }
  
      rm(tD.invR.D)
    }
    #### END #### Iterations##########################################
    if(sr.2>=0.999){print(paste("Warning: estimate for sr.2 out of range:",sr.2,sep=" ")); sr.2 = 0.999; print(c(est.theta=est.theta)); cat(c("\n","\n"))}
    if(sr.2<0){print(paste("Warning: estimate for sr.2 out of range:",sr.2,sep=" ")); sr.2 = 0; print(c(est.theta=est.theta)); cat(c("\n","\n"))}
    if(sr.1>=0.999){print(paste("Warning: estimate for sr.1 out of range:",sr.1,sep=" ")); sr.1 = 0.999; print(c(est.theta=est.theta)); cat(c("\n","\n"))}
    if(sr.1<0){print(paste("Warning: estimate for sr.1 out of range:",sr.1,sep=" ")); sr.1 = 0; print(c(est.theta=est.theta)); cat(c("\n","\n"))}

    if(iter>=max.it | krit!=1){print(paste("Warning: Iterative GEE est. did not converge in simulation:",s,sep=" "))}
  
    est.corr = (est.corr*(s-1) + .5*log((1+c(sr.2,sr.1))/(1-c(sr.2,sr.1))))/s  # collect estimated correlations
    est.var = (est.var*(s-1) + variance)/s # collect estimated variances
  
    ### collect estimation results from iterative GEE estimator
    out.iter.v = (out.iter.v*(s-1) + diag(tmp.v.1))/s # estimated variance
    out.iter.b = rbind(out.iter.b,t(est.theta.1))     # estimated theta
    cover.iter.b = (cover.iter.b * (s-1) + (((est.theta.1-sqrt(diag(tmp.v.1))*qnorm(1-alpha/2)) < theta) & (theta < (est.theta.1-sqrt(diag(tmp.v.1))*qnorm(alpha/2)))))/s
    # actual coverage of two-sided symmetric confidence itervals assuming normality 
    rm(est.theta.1,tmp.v.1)
  } # End: Iterative GEE 
  # END Estimation of standard GEE estimator with simplifications, see Section 3.1   
  ###################################################################################

  
  ########################################################################################################  
  ###### Bootstrap-Part
  ########################################################################################################    

  mean.y = tapply(y,list(indx2,id),mean) # dim: K x n

  ### collect results for bootstrap confidence intervals, see section 4
  boot.Wald = NULL
  boot.Wald[paste("boot.Wald",names(C.mat))] = list(NULL)

  boot.std = c() # collect recentered elementwise standardized bootstrap estimates of theta using the simple GEE
  boot.lin.b = NULL
  boot.lin.b[paste("boot.lin.b",names(C.mat))] = list(NULL)

  temp.KI.b = c()
  # start bootstrap loop; cf. algorithm in section 4
  for(b in seq_len(n.boot1)){
    boot.m.y  = t(apply(mean.y,1,function(x) x[sample(n,size=n,replace=TRUE)])) # naive bootstrap 
    boot.theta = invZ%*%rowMeans(boot.m.y) # calculate bootstrap version of simple GEE estimator

    boot.tmp.v = boot.m.y - matrix(Z%*%boot.theta,ncol(Z),n) # bootstrap residuals
    boot.tmp.v = boot.tmp.v%*%t(boot.tmp.v)/(n-1)    # (K x K) variance matrix of residuals
    boot.tmp.v = invZ%*%boot.tmp.v%*%t(invZ)/n # (K x K) (estimated) variance matrix of bootstrap theta estimate

    # collect individual recentered/standardized bootstrap estimates, assuming independence
    boot.std = rbind(boot.std,t((boot.theta-est.theta))/sqrt(diag(boot.tmp.v)) ) 

    boot.W = lapply(seq_len(length(C.mat)),function(x){
                    C.mat[[x]]%*%boot.tmp.v%*%t(C.mat[[x]])}) # calculate "denominator" of bootstrap Wald statistic T_W
    
    # for linear hypotheses, independence assumption about linear combinations of bootstrap theta estimates
    # collect linear combinations standardized of recentered bootstrap estimates, assuming independence
    boot.lin.b = lapply(seq_len(length(C.mat)),function(x){ 
       rbind(boot.lin.b[[x]],t(C.mat[[x]]%*%(boot.theta - est.theta))/sqrt(diag(boot.W[[x]])))}) # see last paragraph section 4

    # for Wald statistics: inverse of C^T \hat{Var}(\hat{theta}) C, generated assuming independence
    # collect bootstrap-Wald statistics
    boot.Wald = lapply(seq_len(length(C.mat)),function(x){
      tmp.W = solve(boot.W[[x]])
      c(boot.Wald[[x]],t(C.mat[[x]]%*%(boot.theta - est.theta))%*%tmp.W%*%(C.mat[[x]]%*%(boot.theta - est.theta)))})
    
    rm(boot.m.y,boot.theta,boot.tmp.v,boot.W)
  }
  # boot.Wald: bootstrap Wald statistics
  # boot.std: bootstrap standardized (under independence) values for theta 
  # boot.lin.b: for linear hypotheses, independence assumption  

  # bootstrap version of Wald statistic
  Wald.b = lapply(seq_len(length(C.mat)),function(x)
           (Wald.b[[x]]*(s-1) + as.numeric(stand.W[[x]] <= quantile(boot.Wald[[x]],probs=1-alpha,type=8)))/s) 
  rm(boot.Wald)
  
  ### Start --- Bias corrected accelerated Bootstrap: BC_a
  # calculate bias corrections, z_0
  if(ncol(boot.std)>1){
    z.std.0 = qnorm(rowMeans(apply(boot.std,1,function(x) as.numeric(x < 0)))) # standardized values
  }else{
    z.std.0 = qnorm(rowMeans(as.numeric(boot.std < 0)))    
  }
  # standardized values, linear hypothesis
  z.lin.0 = lapply(seq_len(length(C.mat)),function(x){
    tmp =  boot.lin.b[[x]]
    if(ncol(tmp)>1){qnorm(rowMeans(apply(boot.lin.b[[x]],1,function(x) as.numeric(x < 0))))}else{qnorm(mean(as.numeric(tmp < 0)))}})

  ## collect bootstrap estimates to calculate the acceleration parameter a (see Efron and Tibshirani, 1993)
  temp.std.bb = c() # collect recentered and individually standardized theta estimates   
  temp.lin.b = NULL # collect results for standardized linear combinations of recentered estimates
  temp.lin.b[paste("temp.lin.b",names(C.mat))] = list(NULL)
  
  for(j in seq_len(n.boot2)){
     boot.m.y  = t(apply(mean.y,1,function(x) x[sample(n,size=n,replace=TRUE)])) # (no of parameters) x n

     temp.b = invZ%*%rowMeans(boot.m.y) # estimae of theta
     temp.v = boot.m.y - matrix(Z%*%temp.b,ncol(Z),ncol(boot.m.y))
     temp.v = temp.v%*%t(temp.v)/(ncol(boot.m.y)-1)    
     temp.v = invZ%*%temp.v%*%t(invZ)/ncol(boot.m.y) # variance estimate

     temp.std.bb = rbind(temp.std.bb,as.vector((temp.b-est.theta)/sqrt(diag(temp.v)))) # standardized individual values

     # for linear hypotheses, hypotheses tests
     temp.lin.b = lapply(seq_len(length(C.mat)),function(x){ 
                         temp.v = C.mat[[x]]%*%temp.v%*%t(C.mat[[x]])
                         temp.b = t(C.mat[[x]]%*%(temp.b-est.theta))/sqrt(diag(temp.v)) # standardized version
                         rbind(temp.lin.b[[x]],temp.b)})
  }
  rm(boot.m.y,temp.b,temp.v)

  # acceleration parameter for BC_a bootstrap
  a = lapply(seq_len(length(C.mat)),function(x) accel(temp.lin.b[[x]])) # linear hypothesis - standardized version
  a.std.bb = accel(temp.std.bb) # standardized single parameters
  
  # Confidence intervals
  # for linear combinations

  ## two-sided confidence intervals, see Section 4
  temp = lapply(seq_len(length(C.mat)),function(x){
    Lo = pnorm(z.lin.0[[x]] + (z.lin.0[[x]] + qnorm(alpha/2))/(1- a[[x]]*(z.lin.0[[x]] + qnorm(alpha/2)))) # standardized version BCA
    Up = pnorm(z.lin.0[[x]] + (z.lin.0[[x]] + qnorm(1-alpha/2))/(1- a[[x]]*(z.lin.0[[x]] + qnorm(1-alpha/2)))) # standardized version BCA
    temp = vapply(1:ncol(boot.lin.b[[x]]),function(y){quantile(boot.lin.b[[x]][,y],probs=c(Lo[y],Up[y]),type=8)},
                  FUN.VALUE=numeric(2))
    Lo = C.mat[[x]]%*%est.theta - sqrt(diag(C.mat[[x]]%*%tmp.v%*%t(C.mat[[x]])))*temp[2,]
    Up = C.mat[[x]]%*%est.theta - sqrt(diag(C.mat[[x]]%*%tmp.v%*%t(C.mat[[x]])))*temp[1,]
    cbind(Lo,Up)
  })
  # standardized version BCA
  LoUp.b = lapply(seq_len(length(C.mat)),
                  function(x)(LoUp.b[[x]]*(s-1)+cbind(temp[[x]][,1],temp[[x]][,2]))/s)
  bca.lin.n.b = lapply(seq_len(length(C.mat)),
                       function(x)(bca.lin.n.b[[x]]*(s-1) + as.numeric(temp[[x]][,1] < test.zeta[[x]] & test.zeta[[x]] < temp[[x]][,2]))/s)
  rm(temp)

  # single parameters: standardized
  Lo = pnorm(z.std.0 + (z.std.0 + qnorm(alpha/2))/(1- a.std.bb*(z.std.0 + qnorm(alpha/2))))
  Up = pnorm(z.std.0 + (z.std.0 + qnorm(1-alpha/2))/(1- a.std.bb*(z.std.0 + qnorm(1-alpha/2))))
  temp = sapply(1:ncol(boot.std),function(x){quantile(boot.std[,x],probs=c(Lo[x],Up[x]),type=8)})
  Lo = est.theta - sqrt(diag(tmp.v))*temp[2,]
  Up = est.theta - sqrt(diag(tmp.v))*temp[1,]
  bca.std.n.b = (bca.std.n.b*(s-1) + as.numeric(Lo < theta & theta < Up))/s

  ## one-sided confidence intervals  
  # linear hypotheses
  temp = lapply(seq_len(length(C.mat)),function(x){
    Lo = pnorm(z.lin.0[[x]] + (z.lin.0[[x]] + qnorm(alpha))/(1- a[[x]]*(z.lin.0[[x]] + qnorm(alpha)))) # standardized version BCA
    Up = pnorm(z.lin.0[[x]] + (z.lin.0[[x]] + qnorm(1-alpha))/(1- a[[x]]*(z.lin.0[[x]] + qnorm(1-alpha)))) # standardized version BCA
    temp = vapply(1:ncol(boot.lin.b[[x]]),function(y){quantile(boot.lin.b[[x]][,y],probs=c(Lo[y],Up[y]),type=8)},
                  FUN.VALUE=numeric(2))
    Lo = C.mat[[x]]%*%est.theta - sqrt(diag(C.mat[[x]]%*%tmp.v%*%t(C.mat[[x]])))*temp[1,]
    Up = C.mat[[x]]%*%est.theta - sqrt(diag(C.mat[[x]]%*%tmp.v%*%t(C.mat[[x]])))*temp[2,]
    cbind(Lo,Up)
  })
  bca.lin.ones.n.b = lapply(seq_len(length(C.mat)),
                            function(x)(bca.lin.ones.n.b[[x]]*(s-1)+cbind(test.zeta[[x]] < temp[[x]][,1],temp[[x]][,2] < test.zeta[[x]]))/s)

  # single parameters: standardized  
  Lo = pnorm(z.std.0 + (z.std.0 + qnorm(alpha))/(1- a.std.bb*(z.std.0 + qnorm(alpha))))
  Up = pnorm(z.std.0 + (z.std.0 + qnorm(1-alpha))/(1- a.std.bb*(z.std.0 + qnorm(1-alpha))))
  temp = sapply(1:ncol(boot.std),function(x){quantile(boot.std[,x],probs=c(Lo[x],Up[x]),type=8)})
  Lo = est.theta - sqrt(diag(tmp.v))*temp[1,] 
  Up = est.theta - sqrt(diag(tmp.v))*temp[2,] 
  bca.std.n.ones.b = (bca.std.n.ones.b*(s-1)+cbind( theta < Lo, Up < theta))/s

  rm(Lo,Up,temp,temp.std.bb,temp.lin.b,a,a.std.bb,z.std.0,z.lin.0)
  
  ####### End --- Bias corrected accelerated Bootstrap: BC_a
  
  
  #####################################################################################
  # the following prints out information during the simulations
  time.neu = Sys.time()
  d.time = time.neu-time.old
  mean.time = (mean.time*(s-1)+d.time)/s
  print(c("nsim"=noquote(nsim),      # number of simulations
          "  sim"=noquote(as.character(round(s,0))),  # current simulation  
          "  min/sim"=noquote(as.character(round(d.time/60,2))), # average of minutes needes per simputation 
          "  remain"=noquote(as.character(round((nsim-s)*mean.time/60,2))), # estimated remaining time 
          "  Wald"=noquote(as.character(round(Wald[[1]],3))),
          "  Wald.b"=noquote(as.character(round(Wald.b[[1]],3)))))
}

####################### End Simulation #########################################################################

####################### Part 4: Output of Results ##############################################################

# --------------------------------------------- Output ---------------------------------------------- 
# results referred to in the paper are uncommented; additional results are available, but corresponding lines need to be uncommented

text = c("Standard Normal",paste("T df=",dist.p["fg"],sep=""),paste("Chisq df=",dist.p["fg"],sep=""),
         paste("Inverse Gauss, mu=",dist.p["mu"],", lambda=",dist.p["lambda"],sep=""),
         paste("Negative inverse Gauss, mu=",dist.p["mu"],", lambda=",dist.p["lambda"],sep=""))

thet.var = noquote(rbind(c("theta",round(summary(theta),4)),c("variances",round(summary(s2.v),4))))

# results for individual elements in theta: simple GEE and simplified iterative GEE

simple.GEE = rbind(round(theta,4),round(colMeans(out.b),4),round(sqrt(out.v),4),
                   round(sqrt(diag(cov(out.b))),4),round(t(out.ci),3),round(t(bca.std.n.b),3))
rownames(simple.GEE) = c("true value","mean estim.","sd of estim.","mean se","cover. N","cover. BCa")

cover.one.sided = rbind(round(t(out.ci.l),3),round(t(out.ci.u),3),round(t(bca.std.n.ones.b),3))
rownames(cover.one.sided) = c("cover. lo-N", "cover. up-N", "cover. lo-ci", "cover. up-ci")

if(GEE.iter==1){
  iter.GEE = rbind(round(theta,4),round(colMeans(out.iter.b),4),round(sqrt(out.iter.v),4),
                     round(sqrt(diag(cov(out.iter.b))),4),round(t(cover.iter.b),3))
  rownames(iter.GEE) = c("true value","mean estim.","sd of estim.","mean se","cover. N")

  corr.est = matrix(round((exp(2*est.corr)-1)/(exp(2*est.corr)+1),4),1,length(est.corr))
  rownames(corr.est) = "Estimated correlations: "

  var.est = matrix(round(est.var,4),1,length(est.var))
  rownames(var.est) = "Estimated variances: "

  comp.simple.iterative = rbind(round(colMeans(out.b)/colMeans(out.iter.b),4),round(sqrt(out.v)/sqrt(out.iter.v),4))
  rownames(comp.simple.iterative) = c("mean(simple)/mean(iter.)","sd(simple)/sd(iter.)")

  trace.vars = noquote(paste("Traces of variance matrices: ","tr(var.simple.GEE)",round(sum(diag(cov(out.b))),4),
                             "tr(var.iterative.GEE)",round(sum(diag(cov(out.iter.b))),4)))
}

# Wald test for linear hypotheses 
#Wald.tests = lapply(seq_len(length(C.mat)),function(x){
#                   print(paste("Actual non-rejection rate Wald tests.  ",names(C.mat)[[x]], 
#                               "Standard Wald test: ",round(c(Wald=Wald[[x]]),4), 
#                               "Bootstrap Wald test: ",round(c(Wald.b=Wald.b[[x]]),4)))}) # bootstrap version

Wald.tests = round(cbind(standard = Reduce(rbind,Wald), Bootstrap = Reduce(rbind,Wald.b)),4)
rownames(Wald.tests) = names(C.mat)
colnames(Wald.tests) = c("Wald","Wald.b")

if(rep.anov==1){
  # Standard Repeated measures ANOVA with Greenhouse-Geisser correction: F-test, actual rejection rates
  anov = round(res.stand,4)
  anov.nam = t(c("Repeated Measures ANOVA results: Actual non-rejection rate:"))
  overall = "Overall test:"
  anov.overall = round(overall.F,4)
} 


############# 
# Output for Tables 1-4
# select the elements for Tables 1 and 2
if(rep.anov==1){selection = round(quantile(seq_len(length(theta)),p=c(0,0.25,0.5,0.75,1)))}

if(rep.anov==0 & length(theta)==32){
  tmp = order(theta)
  selection = tmp[1]
  selection = c(selection,tmp[9])
  tmp.1 = order(abs(c(theta[tmp[16]],theta[tmp[17]])))[1]
  selection = c(selection,c(tmp[16],tmp[17])[tmp.1]) 
  selection = c(selection,tmp[24])
  selection = c(selection,tmp[length(theta)])
}

print.out = c()
for(a in selection){
  temp = paste(c(simple.GEE[1:3,a],simple.GEE[5:6,a]),collapse=" & ")
  print.out = rbind(print.out,temp)
}
rownames(print.out) = rep("",nrow(print.out))
colnames(print.out) = c("true & est & sd & est.iter & cov-N & cov-BCA")
print.out = noquote(print.out)

########################### Now print output ##########################################################
cat("No of units: ",n,"\n",
    "No. of simulations: ", nsim, "\n",
    "Distribution of random effects: ","\n", text[dist.re[1]+1],"\n",
    text[dist.re[2]+1],"\n",text[dist.re[3]+1],"\n\n",
    "Skewness of error: ",round(skew,4),"\n",
    "Excess kurtosis of error: ",round(kurt-3,4))  

ifelse(length(s2.v)==n*prod(K),print("Variances vary over units"),print("Variances do not vary over units"))

thet.var
cat("rho.2: ",rho.2,"rho.1: ",rho.1,"\n")

cat("times: ",times,"\n")

noquote(rbind(c("p.T.ik: ",round(summary(p.T.ik),4)))) 

noquote(c("No. of bootstrap samples",n.boot1=n.boot1,n.boot2=n.boot2))

# results for individual elements in theta: simple GEE and simplified iterative GEE
simple.GEE 
#cover.one.sided 
if(GEE.iter==1){iter.GEE}
if(GEE.iter==1){comp.simple.iterative }
if(GEE.iter==1){corr.est} 
if(GEE.iter==1){var.est}
if(GEE.iter==1){trace.vars}

# Standard Repeated measures ANOVA with Greenhouse-Geisser correction: F-test, actual rejection rates
# makes only sense if theta = 0
if(rep.anov==1){anov.nam}
if(rep.anov==1){anov}
if(rep.anov==1){overall}
if(rep.anov==1){anov.overall}

# confidence intervals for linear combinations - not part of the paper
linear.comb = lapply(seq_len(length(C.mat)),function(x){
  tmp = t(round(cbind(zeta[[x]],est.zeta[[x]],test.zeta[[x]],out.ci.a[[x]],bca.lin.n.b[[x]],ci.LoUp.a[[x]],bca.lin.ones.n.b[[x]]),4))
  rownames(tmp)= c("true zeta","estimated zeta","tested zeta","NV 2-sided","BCA 2-sided",
                   "NV lower 1-sided", "NV upper 1-sided",
                   "BCA lower 1-sided", "BCA upper 1-sided")
  return(tmp)  
})
names(linear.comb) = names(C.mat)
linear.comb

# confidence interval bounds for linear combinations - not part of the paper
lincomb.ci.width = lapply(seq_len(length(C.mat)),function(x){
  tmp = t(round(cbind(zeta[[x]],est.zeta[[x]],test.zeta[[x]],LoUp.a[[x]],LoUp.b[[x]]),4)) 
  rownames(tmp)= c("true zeta","estimated zeta","tested zeta","NV 2-sided lower bound", "NV 2-sided upper bound",
                   "BCA 2-sided lower bound", "BCA 2-sided upper bound")
  return(tmp)  
})
names(lincomb.ci.width) = names(C.mat)
lincomb.ci.width

## test.zeta
cat("test.zeta","\n")
lapply(test.zeta,function(x) t(x))

# Wald test for linear hypothesis 
Wald.tests

# Simple GEE estimator for Tables: 
print.out
########################### End print output ##########################################################
